#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2024 Red Hat

from ctypes import (
    c_uint32,
    c_int32,
    c_void_p,
)
from typing import Optional, Tuple, Type, Self
from dataclasses import dataclass
from enum import IntEnum

import ctypes
import pytest
import types


# see include/linux/uapi/btf.h
# struct btf_type {
#         __u32 name_off;
#         /* "info" bits arrangement
#          * bits  0-15: vlen (e.g. # of struct's members)
#          * bits 16-23: unused
#          * bits 24-28: kind (e.g. int, ptr, array...etc)
#          * bits 29-30: unused
#          * bit     31: kind_flag, currently used by
#          *             struct, union, enum, fwd and enum64
#          */
#         __u32 info;
#         /* "size" is used by INT, ENUM, STRUCT, UNION, DATASEC and ENUM64.
#          * "size" tells the size of the type it is describing.
#          *
#          * "type" is used by PTR, TYPEDEF, VOLATILE, CONST, RESTRICT,
#          * FUNC, FUNC_PROTO, VAR, DECL_TAG and TYPE_TAG.
#          * "type" is a type_id referring to another type.
#          */
#         union {
#                 __u32 size;
#                 __u32 type;
#         };
# };
class _U(ctypes.Union):
    _fields_ = [
        ("size", c_uint32),
        ("type", c_uint32),
    ]


class BtfType(ctypes.Structure):
    _anonymous_ = ("u",)
    _fields_ = [
        ("name_off", c_uint32),
        ("info", c_uint32),
        ("u", _U),
    ]

    @property
    def vlen(self):
        """#define BTF_INFO_VLEN(info)     ((info) & 0xffff)"""
        return self.info & 0xFFFF

    @property
    def kind(self):
        """#define BTF_INFO_KIND(info)     (((info) >> 24) & 0x1f)"""
        return (self.info >> 24) & 0x1F

    @property
    def kflag(self):
        """#define BTF_INFO_KFLAG(info)    ((info) >> 31)"""
        return self.info >> 31


class BtfKind(IntEnum):
    UNKN = 0  # Unknown
    INT = 1  # Integer
    PTR = 2  # Pointer
    ARRAY = 3  # Array
    STRUCT = 4  # Struct
    UNION = 5  # Union
    ENUM = 6  # Enumeration up to 32-bit values
    FWD = 7  # Forward
    TYPEDEF = 8  # Typedef
    VOLATILE = 9  # Volatile
    CONST = 10  # Const
    RESTRICT = 11  # Restrict
    FUNC = 12  # Function
    FUNC_PROTO = 13  # Function Proto
    VAR = 14  # Variable
    DATASEC = 15  # Section
    FLOAT = 16  # Floating point
    DECL_TAG = 17  # Decl Tag
    TYPE_TAG = 18  # Type Tag
    ENUM64 = 19  # Enumeration up to 64-bit values


# /* BTF_KIND_STRUCT and BTF_KIND_UNION are followed
#  * by multiple "struct btf_member".  The exact number
#  * of btf_member is stored in the vlen (of the info in
#  * "struct btf_type").
#  */
# struct btf_member {
#         __u32   name_off;
#         __u32   type;
#         /* If the type info kind_flag is set, the btf_member offset
#          * contains both member bitfield size and bit offset. The
#          * bitfield size is set for bitfield members. If the type
#          * info kind_flag is not set, the offset contains only bit
#          * offset.
#          */
#         __u32   offset;
# };
class BtfMember(ctypes.Structure):
    _fields_ = [
        ("name_off", c_uint32),
        ("type", c_uint32),
        ("offset", c_uint32),
    ]

    # /* If the struct/union type info kind_flag is set, the
    #  * following two macros are used to access bitfield_size
    #  * and bit_offset from btf_member.offset.
    #  */
    @property
    def bitfield_size(self):
        return self.offset >> 24

    @property
    def bit_offset(self):
        return self.offset & 0xFFFFFF


# /* BTF_KIND_ARRAY is followed by one "struct btf_array" */
# struct btf_array {
# 	__u32	type;
# 	__u32	index_type;
# 	__u32	nelems;
# };
class BtfArray(ctypes.Structure):
    _fields_ = [
        ("type", c_uint32),
        ("index_type", c_uint32),
        ("nelems", c_uint32),
    ]


# /* BTF_KIND_FUNC_PROTO is followed by multiple "struct btf_param".
#  * The exact number of btf_param is stored in the vlen (of the
#  * info in "struct btf_type").
#  */
# struct btf_param {
# 	__u32	name_off;
# 	__u32	type;
# };
class BtfParam(ctypes.Structure):
    _fields_ = [
        ("name_off", c_uint32),
        ("type", c_uint32),
    ]


# /* BTF_KIND_DATASEC is followed by multiple "struct btf_var_secinfo"
#  * to describe all BTF_KIND_VAR types it contains along with it's
#  * in-section offset as well as size.
#  */
# struct btf_var_secinfo {
# 	__u32	type;
# 	__u32	offset;
# 	__u32	size;
# };
class BtfVarSecinfo(ctypes.Structure):
    _fields_ = [
        ("type", c_uint32),
        ("offset", c_uint32),
        ("size", c_uint32),
    ]


@dataclass
class Api:
    """
    Wrapper to make automatically loading functions from a .so file simpler.
    """

    name: str
    args: Tuple[Type[ctypes._SimpleCData | ctypes._Pointer], ...]
    return_type: Optional[Type[ctypes._SimpleCData | ctypes._Pointer]]


@dataclass
class Map:
    name: str
    ctype: type


class Btf:
    # Cached libbpf .so file
    _lib: ctypes.CDLL | None = None

    # Cached per .so btf files
    _btfs: dict[str, Self] = {}

    _btf_api_prototypes: list[Api] = [
        Api(
            name="btf__load_vmlinux_btf",
            args=(),
            return_type=c_void_p,
        ),
        Api(
            name="btf__new_empty",
            args=(),
            return_type=c_void_p,
        ),
        Api(
            name="btf__parse",
            args=(ctypes.c_char_p, c_void_p),
            return_type=c_void_p,
        ),
        Api(
            name="btf__add_btf",
            args=(c_void_p, c_void_p),
            return_type=c_int32,
        ),
        Api(
            name="btf__find_by_name",
            args=(c_void_p, ctypes.c_char_p),
            return_type=c_int32,
        ),
        Api(
            name="btf__type_by_id",
            args=(c_void_p, c_uint32),
            return_type=c_void_p,
        ),
        Api(
            name="btf__name_by_offset",
            args=(c_void_p, c_uint32),
            return_type=ctypes.c_char_p,
        ),
    ]

    def __init__(self, name: str, lib: ctypes.CDLL, btf: c_void_p):
        self.lib = lib
        self.name = name
        self.btf = btf
        self.known_types: dict[str, ctypes.Structure | ctypes.Union] = {}
        self.ignore_types: list[ctypes.Structure | ctypes.Union] = []
        Btf._btfs[name] = self

    @classmethod
    def load(cls, btf_files: None | list[str] = None) -> Self:
        """
        Load BTF from the running kernel
        """
        if cls._lib is None:
            cls._lib = cls._load_libbpf(btf_files)

        name = btf_files[0].parent.stem if btf_files is not None else "vmlinux"

        instance = cls._btfs.get(name, cls._load(name, btf_files))
        assert instance is not None
        return instance

    @classmethod
    def _load_libbpf(cls, btf_files: None | list[str]) -> Self:
        try:
            libbpf = ctypes.CDLL("libbpf.so", use_errno=True)
            assert libbpf is not None
        except OSError as e:
            pytest.exit(
                f"Error loading the library: {e}. Maybe libbpf is not installed?"
            )
        for api in Btf._btf_api_prototypes:
            fun = getattr(libbpf, api.name)
            fun.argtypes = api.args
            fun.restype = api.return_type
        return libbpf

    @classmethod
    def _load(cls, name: str, btf_files: None | list[str]) -> Self:
        libbpf = cls._lib
        if btf_files is None:
            _btf = c_void_p(libbpf.btf__load_vmlinux_btf())
        else:
            _btf = c_void_p(libbpf.btf__new_empty())
            for btf_file in btf_files:
                btf = c_void_p(libbpf.btf__parse(bytes(btf_file), c_void_p(None)))
                err = libbpf.btf__add_btf(_btf, btf)
                assert err >= 0

        assert _btf.value is not None
        return cls(name, libbpf, _btf)

    def get_type(self, name, type_id, from_pointer=False):
        btf = self.btf
        libbpf = self.lib
        btf_to_ctypes = {
            b"_Bool": ctypes.c_bool,
            b"char": ctypes.c_char,
            b"signed char": ctypes.c_char,
            b"unsigned char": ctypes.c_ubyte,
            b"short int": ctypes.c_short,
            b"short unsigned int": ctypes.c_ushort,
            b"int": ctypes.c_int,
            b"unsigned int": ctypes.c_uint,
            b"long int": ctypes.c_long,
            b"long unsigned int": ctypes.c_ulong,
            b"long long int": ctypes.c_longlong,
            b"long long unsigned int": ctypes.c_ulonglong,
            b"__int128 unsigned": ctypes.c_longlong * 2,
            b"size_t": ctypes.c_size_t,
        }

        m_type_p = libbpf.btf__type_by_id(btf, type_id)
        if not m_type_p:
            return None

        m_type = BtfType.from_address(m_type_p)
        m_type_name = libbpf.btf__name_by_offset(btf, m_type.name_off)
        kind = m_type.kind

        if kind == BtfKind.UNKN:
            return None

        elif kind == BtfKind.INT:
            # basic type in C
            return btf_to_ctypes[m_type_name]

        elif kind == BtfKind.PTR:
            # pointer: if we get a new type behind that pointer, the
            # new type will not be recursively populated
            val_type = self.get_type(name, m_type.type, from_pointer=True)
            fun_type = ctypes.CFUNCTYPE(None)
            if val_type is None:
                return ctypes.c_void_p
            elif isinstance(val_type, type(fun_type)):
                return val_type
            return ctypes.POINTER(val_type)

        elif kind == BtfKind.ARRAY:
            # int[length]
            array_info = BtfArray.from_address(m_type_p + ctypes.sizeof(BtfType))
            return self.get_type(name, array_info.type) * array_info.nelems

        elif kind in [BtfKind.STRUCT, BtfKind.UNION]:
            # recursively go down the struct or union if we are not coming
            # from a pointer
            name = m_type_name.decode()
            if not name:
                name = f"anon_{type_id}"

            cls = self.known_types.get(name, None)
            if cls is not None and hasattr(cls, "_fields_"):
                # the type is already fully known, return
                return cls

            elif cls is None:
                # new struct/union, allocate a new class
                parent = ctypes.Structure if kind == BtfKind.STRUCT else ctypes.Union
                camelCaseName = "".join([word.capitalize() for word in name.split("_")])
                cls = types.new_class(camelCaseName, (parent,))
                cls.cname = name
                self.known_types[name] = cls

            if from_pointer:
                return cls

            MEMBERS = BtfMember * m_type.vlen
            arg_members = MEMBERS.from_address(m_type_p + ctypes.sizeof(BtfType))

            fields = []
            anonymous = []
            unique_id = 0
            kind_flag = m_type.kflag
            for member in arg_members:
                name = libbpf.btf__name_by_offset(btf, member.name_off).decode()
                ctype = None

                if not name:
                    # if we have an anonymous field, register it
                    name = f"u_{unique_id}"
                    unique_id += 1
                    ctype = self.get_type(name, member.type)
                    anonymous.append(name)
                elif (
                    hasattr(cls, "_fields_overrides_")
                    and name in cls._fields_overrides_
                ):
                    ctype = cls._fields_overrides_[name]
                else:
                    ctype = self.get_type(name, member.type)

                assert ctype is not None

                # some ints have a bitfield size, ensure we don't mess up
                if not kind_flag or not member.bitfield_size:
                    fields.append((name, ctype))
                else:
                    fields.append((name, ctype, member.bitfield_size))

            if anonymous:
                cls._anonymous_ = anonymous
            cls._fields_ = fields

            return cls

        elif kind == BtfKind.ENUM:
            # enums are translated to int, long, etc based on the size parameter
            size = m_type.size * 8
            return getattr(ctypes, f"c_uint{size}")

        elif kind == BtfKind.FWD:
            # Forward are types that are not loaded yet, so we can't
            # type them
            return None

        elif kind == BtfKind.TYPEDEF:
            # typedefs are like int, but we might have specifics in ctypes,
            # so look in btf_to_ctypes first.
            return btf_to_ctypes.get(
                m_type_name, self.get_type(m_type_name, m_type.type)
            )

        elif kind in [BtfKind.VOLATILE, BtfKind.CONST]:
            # modifier, just forward the child type
            return self.get_type(name, m_type.type, from_pointer)

        elif kind == BtfKind.FUNC_PROTO:
            ret_type = self.get_type(name, m_type.type)
            ARGS = BtfParam * m_type.vlen
            args = ARGS.from_address(m_type_p + ctypes.sizeof(BtfType))
            params = [
                self.get_type(libbpf.btf__name_by_offset(btf, a.name_off), a.type)
                for a in args
            ]
            # special case `int fun(void)`
            if len(params) == 1 and params[0] is None:
                params = []
            return ctypes.CFUNCTYPE(ret_type, *params)

        else:
            raise Exception(f"unsupported btf kind {kind} for {name}")

    def build_struct(self, cls):
        """
        Take the given ctypes class, look for its definition in the btf
        and populate its _fields_ field.

        Note that this allows forward declaration of pointers, thus if
        a struct is expected to have a pointer and we need to dereference
        it, the pointed struct must be build first.
        """
        btf = self.btf
        libbpf = self.lib
        cname = cls.cname

        if cname not in self.known_types:
            self.known_types[cname] = cls

        if hasattr(cls, "_fields_"):
            return

        self.get_type(cname, libbpf.btf__find_by_name(btf, cname.encode()))

    @property
    def maps(self):
        btf = self.btf
        libbpf = self.lib
        type_id = libbpf.btf__find_by_name(btf, ".maps".encode())
        if type_id < 0:
            return []

        m_type_p = libbpf.btf__type_by_id(btf, type_id)
        m_type = BtfType.from_address(m_type_p)
        kind = m_type.kind

        assert kind == BtfKind.DATASEC

        if m_type.vlen == 0:
            return []

        MAPS = BtfVarSecinfo * m_type.vlen
        maps = MAPS.from_address(m_type_p + ctypes.sizeof(BtfType))

        maps_types = [
            BtfType.from_address(libbpf.btf__type_by_id(btf, m.type)) for m in maps
        ]

        retval = []

        for m in maps_types:
            map_name = libbpf.btf__name_by_offset(btf, m.name_off).decode()
            type_id = m.type

            ctype = self.get_type("", type_id)
            value = [f[1] for f in ctype._fields_ if f[0] == "value"][0]

            retval.append(Map(map_name, value._type_))

        return retval



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2024 Red Hat

from . import Bpf, HidProbeArgs, PrivateTestData

import binascii
import logging
import struct
import pytest

from dataclasses import dataclass

logger = logging.getLogger(__name__)


@pytest.fixture
def bpf(source: str):
    """
    A fixture that allows parametrizing over a number of sources. Use with e.g.::

        @pytest.mark.parametrize("source", ["0010-FR-TEC__Raptor-Mach-2"])
        def test_something(bpf):
            pass
    """
    assert source is not None
    bpf = Bpf.load(source)
    assert bpf is not None
    yield bpf


@pytest.mark.parametrize("source", ["0010-FR-TEC__Raptor-Mach-2"])
class TestFrTecRaptorMach2:
    def test_probe(self, bpf):
        probe_args = HidProbeArgs(rdesc_size=232)
        probe_args.rdesc[177] = 0xEF
        pa = bpf.probe(probe_args)
        assert pa.retval == 0

        probe_args.rdesc[177] = 0x12  # random value
        pa = bpf.probe(probe_args)
        assert pa.retval == -22

    def test_rdesc(self, bpf):
        rdesc = bytes(4096)

        data = bpf.hid_bpf_rdesc_fixup(rdesc=rdesc)
        assert data[177] == 0x07


@pytest.mark.parametrize("source", ["0010-mouse_invert_y"])
class TestUserhacksInvertY:
    def test_probe(self, bpf):
        probe_args = HidProbeArgs()
        probe_args.rdesc_size = 123
        pa = bpf.probe(probe_args)
        assert pa.retval == -22

        probe_args.rdesc_size = 71
        pa = bpf.probe(probe_args)
        assert pa.retval == 0

    @pytest.mark.parametrize("y", [1, -1, 10, -256])
    def test_event(self, bpf, y):
        # this device has reports of size 9
        values = (0, 0, 0, y, 0, 0, 0, 0, 0)
        report = struct.pack("<3bh5b", *values)

        values = bpf.hid_bpf_device_event(report=report)
        values = struct.unpack("<3bh5b", values)
        y_out = values[3]
        assert y_out == -y


@pytest.mark.parametrize("source", ["0010-Rapoo__M50-Plus-Silent"])
class TestRapooM50Plus:
    def test_rdesc_fixup(self, bpf):
        rdesc = bytearray(4096)
        rdesc[17] = 0x03

        data = bpf.hid_bpf_rdesc_fixup(rdesc=rdesc)
        rdesc[17] = 0x05
        assert data == rdesc


@pytest.mark.parametrize("source", ["0010-XPPen__DecoMini4"])
class TestXPPenDecoMini4:
    @pytest.mark.parametrize(
        "report,expected",
        [
            # Invalid report descriptor
            ("02 01 02 03 04 05 06 07", "02 01 02 03 04 05 06 07"),
            # Button 1
            ("06 00 05 00 00 00 00 00", "06 01 00 00 00 00 00 00"),
            # Button 2
            ("06 00 08 00 00 00 00 00", "06 02 00 00 00 00 00 00"),
            # Button 3
            ("06 04 00 00 00 00 00 00", "06 04 00 00 00 00 00 00"),
            # Button 4
            ("06 00 2c 00 00 00 00 00", "06 08 00 00 00 00 00 00"),
            # Button 5
            ("06 01 16 00 00 00 00 00", "06 10 00 00 00 00 00 00"),
            # Button 6
            ("06 01 1d 00 00 00 00 00", "06 20 00 00 00 00 00 00"),
            # Buttons 3 and 5
            ("06 05 16 00 00 00 00 00", "06 14 00 00 00 00 00 00"),
            # All buttons
            ("06 05 05 08 2c 16 1d 00", "06 3f 00 00 00 00 00 00"),
        ],
    )
    def test_button_events(self, bpf, report, expected):
        event = bpf.hid_bpf_device_event(report=report)
        assert event == expected


@pytest.mark.parametrize("source", ["0010-TUXEDO__Sirius-16-Gen1-and-Gen2"])
class TestTUXEDOSirius16Gen1andGen2:
    @pytest.mark.parametrize(
        "report,expected",
        [
            pytest.param(
                "01 00 00 68 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00",
                "01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00",
                id="single-f13-key-press",
            ),
            pytest.param(
                "01 00 00 04 05 06 07 08 09 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00",
                "01 00 00 04 05 06 07 08 09 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00",
                id="six-keys-and-then-f13-key-down",
            ),
            pytest.param(
                "01 00 00 68 68 68 68 68 68 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff",
                "01 00 00 00 00 00 00 00 00 ff ff ff ff ff ff ff ff ff ff ff ff ff fe ff ff ff ff ff ff ff ff ff ff ff ff ff ff",
                id="edge-case-all-bits-set",
            ),
        ],
    )
    def test_button_events(self, bpf, report, expected):
        event = bpf.hid_bpf_device_event(report=report)
        assert event == expected


@pytest.mark.parametrize("source", ["0010-XPPen__ACK05"])
class TestXPPenACK05:
    @pytest.mark.parametrize(
        "report,expected",
        [
            # anything but report ID 02 should be forwarded as such
            pytest.param(
                "04 f0 01 00 00 00 00 00 00 00 00 00",
                "04 f0 01 00 00 00 00 00 00 00 00 00",
                id="untouched-report",
            ),
            # button events are on report ID f0 with a size of 8, mostly untouched
            pytest.param(
                "02 f0 01 00 00 00 00 00 00 00 00 00",
                "f0 f0 01 00 00 00 00 00",
                id="single-button1-press",
            ),
            # CCW wheel events are button events with 0x02 changed to 0xff
            pytest.param(
                "02 f0 00 00 00 00 00 02 00 00 00 00",
                "f0 f0 00 00 00 00 00 ff",
                id="ccw-wheel",
            ),
            # battery reports are on f2 and of size 5
            pytest.param(
                "02 f2 01 00 00 00 00 00 00 00 00 00",
                "f2 f2 01 00 00",
                id="battery-report",
            ),
        ],
    )
    def test_button_events(self, bpf, report: str, expected: str):
        event = bpf.hid_bpf_device_event(report=report)
        assert event == expected

    def test_probe(self, bpf):
        private = PrivateTestData(bpf)
        expected_output_data = binascii.unhexlify(
            "02 b0 04 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00".replace(
                " ", ""
            )
        )

        probe_args = HidProbeArgs()
        probe_args.rdesc_size = 37
        pa = bpf.probe(probe_args, private)
        assert pa.retval == -22
        assert len(private.output_reports) == 0

        probe_args.rdesc_size = 36
        pa = bpf.probe(probe_args, private)
        assert pa.retval == 0
        assert len(private.output_reports) == 1
        output_report = private.output_reports.pop()

        assert output_report.data == expected_output_data
        assert output_report.time == 0

    def test_delayed_callback(self, bpf):
        private = PrivateTestData(bpf)
        report = binascii.unhexlify(
            "02 f8 02 01 00 00 00 00 00 00 00 00".replace(" ", "")
        )
        expected_output_data = binascii.unhexlify(
            "02 b0 04 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00".replace(
                " ", ""
            )
        )

        # we need to call .probe() first to initialize the asyncs
        probe_args = HidProbeArgs()
        probe_args.rdesc_size = 36
        pa = bpf.probe(probe_args, private)
        assert pa.retval == 0
        assert len(private.output_reports) == 1
        output_report = private.output_reports.pop()

        assert output_report.data == expected_output_data
        assert output_report.time == 0

        event = bpf.hid_bpf_device_event(report, private)
        assert event == report

        assert len(private.output_reports) == 1
        output_report = private.output_reports.pop()

        assert output_report.data == expected_output_data
        assert output_report.time == 10


@pytest.mark.parametrize("source", ["0020-Huion__Inspiroy-2-S"])
class TestHuionInspiroy2S:
    @pytest.mark.parametrize("down_on_button", [True, False])
    @pytest.mark.parametrize("up_event_count", [1, 2, 3, 4])
    def test_eraser_tip_changes(self, bpf, down_on_button, up_event_count):
        @dataclass
        class Report:
            tip_switch: bool
            secondary_barrel_switch: bool
            in_range: bool = True

            def __bytes__(self) -> bytes:
                id = 0x8
                b1 = 0x0
                if self.in_range:
                    b1 |= 0x80
                if self.tip_switch:
                    b1 |= 0x1
                if self.secondary_barrel_switch:
                    b1 |= 0x4

                # For this test x/y values and pressure don't matter
                return bytes(
                    [id, b1, 0x59, 0x44, 0x47, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
                )

            @classmethod
            def from_bytes(cls, bs: bytes) -> "Report":
                assert bs[0] == 0x8  # Report ID
                return cls(
                    tip_switch=(bs[1] & 0x1) != 0,
                    secondary_barrel_switch=(bs[1] & 0x4) != 0,
                    in_range=(bs[1] & 0x80) != 0,
                )

        # A "random" number of events after our button toggles
        filler_event_count = 10

        reports = [
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=True, secondary_barrel_switch=False),
            # Legitimate tip up/down
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=True, secondary_barrel_switch=False),
            # Press button
            Report(tip_switch=down_on_button, secondary_barrel_switch=True),
            *[Report(tip_switch=False, secondary_barrel_switch=True)] * up_event_count,
            *[Report(tip_switch=True, secondary_barrel_switch=True)] * filler_event_count,
            # Legitimate tip up/down
            Report(tip_switch=False, secondary_barrel_switch=True),
            Report(tip_switch=True, secondary_barrel_switch=True),
            *[Report(tip_switch=True, secondary_barrel_switch=True)] * filler_event_count,
            # Legitimate tip up/down
            # Release button
            Report(tip_switch=down_on_button, secondary_barrel_switch=False),
            *[Report(tip_switch=False, secondary_barrel_switch=False)] * up_event_count,
            *[Report(tip_switch=True, secondary_barrel_switch=False)] * filler_event_count,
            # Legitimate tip up event + prox-out
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=False, secondary_barrel_switch=False, in_range=False),
        ]  # fmt: skip

        expected = [
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=True, secondary_barrel_switch=False),
            # Legitimate tip up/down
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=True, secondary_barrel_switch=False),
            # Press button
            Report(tip_switch=True, secondary_barrel_switch=True),
            *[Report(tip_switch=True, secondary_barrel_switch=True)] * up_event_count,
            *[Report(tip_switch=True, secondary_barrel_switch=True)] * filler_event_count,
            # Legitimate tip up/down
            Report(tip_switch=False, secondary_barrel_switch=True),
            Report(tip_switch=True, secondary_barrel_switch=True),
            *[Report(tip_switch=True, secondary_barrel_switch=True)] * filler_event_count,
            # Release button
            Report(tip_switch=True, secondary_barrel_switch=False),
            *[Report(tip_switch=True, secondary_barrel_switch=False)] * up_event_count,
            *[Report(tip_switch=True, secondary_barrel_switch=False)] * filler_event_count,
            # Legitimate tip up event + prox-out
            Report(tip_switch=False, secondary_barrel_switch=False),
            Report(tip_switch=False, secondary_barrel_switch=False, in_range=False),
        ]  # fmt: skip

        events = [bpf.hid_bpf_device_event(report=bytes(report)) for report in reports]
        events = [Report.from_bytes(e) for e in events]

        assert events == expected



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2024 Red Hat

from ctypes import (
    c_int,
)
from typing import Optional, Tuple, Type, Self
from dataclasses import dataclass
from pathlib import Path
from enum import IntEnum

import binascii
import logging
import ctypes
import os
import json
import pytest
import random
import dataclasses
import errno

from .btf import Btf, Map

logger = logging.getLogger(__name__)
random.seed()


# to be automatically field by BTF thanks to libbpf
class HidDevice(ctypes.Structure):
    cname = "hid_device"


# see struct hid_probe_args
class HidProbeArgs(ctypes.Structure):
    cname = "hid_bpf_probe_args"


# see struct hid_bpf_ctx
class HidBpfCtx(ctypes.Structure):
    cname = "hid_bpf_ctx"


class BpfTimer(ctypes.Structure):
    cname = "bpf_timer"


class BpfWq(ctypes.Structure):
    cname = "bpf_wq"


class TestAsyncCb(ctypes.Structure):
    cname = "test_async_cb"
    _fields_overrides_ = {
        "cb": ctypes.c_void_p,  # just too complex to handle function pointers
    }


@dataclass
class OutputReport:
    time: int
    data: Tuple[bytes]


class ReportType(IntEnum):
    INPUT = 0
    OUTPUT = 1
    FEATURE = 2


class RequestType(IntEnum):
    GET_REPORT = 0x01
    GET_IDLE = 0x02
    GET_PROTOCOL = 0x03
    SET_REPORT = 0x09
    SET_IDLE = 0x0A
    SET_PROTOCOL = 0x0B


@dataclass
class HidRawRequest:
    time: int
    req_data: Tuple[bytes]
    out_data: Tuple[bytes]
    report_type: ReportType
    request_type: RequestType


@dataclass
class PrivateTestData:
    bpf: "Bpf"
    current_ctx: HidBpfCtx = dataclasses.field(default_factory=HidBpfCtx)
    id: int = dataclasses.field(default_factory=lambda: random.randint(0, 0xFFFF))
    output_reports: list[OutputReport] = dataclasses.field(default_factory=list)
    hw_requests: list[HidRawRequest] = dataclasses.field(default_factory=list)
    maps_data: dict[int, dict[int, ...]] = dataclasses.field(default_factory=dict)
    asyncs: dict[int, TestAsyncCb] = dataclasses.field(default_factory=dict)


# see struct test_callbacks
class Callbacks(ctypes.Structure):
    cname = "test_callbacks"
    _fields_overrides_ = {
        "private_data": ctypes.py_object,
    }

    def __init__(self, private: PrivateTestData):
        hid = HidDevice(id=private.id)
        private.current_ctx.hid = ctypes.pointer(hid)

        super().__init__(private_data=ctypes.py_object(private))
        fun_type = ctypes.CFUNCTYPE(None)
        for field, argtype in self._fields_:
            if isinstance(argtype, type(fun_type)):
                fun = getattr(Callbacks, f"_{field}")
                setattr(self, field, argtype(fun))

    def _hid_bpf_allocate_context(callbacks_p, hid: int):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data
        assert pdata.id == hid

        callbacks.ctx = ctypes.pointer(pdata.current_ctx)

        return 0

    def _hid_bpf_release_context(callbacks_p, ctx):
        callbacks = callbacks_p.contents

        callbacks.ctx = None

    @classmethod
    def validate_ctx(cls, callbacks_p, ctx_p):
        callbacks = callbacks_p.contents
        hid = ctx_p.contents.hid.contents
        return hid.id == callbacks.private_data.id

    def _hid_bpf_hw_output_report(callbacks_p, ctx_p, data_p, size):
        if not Callbacks.validate_ctx(callbacks_p, ctx_p):
            return -errno.EINVAL
        DataArray = ctypes.c_uint8 * size
        c_data = DataArray()
        p2 = ctypes.byref(c_data)
        ctypes.memmove(p2, data_p, size)
        data = bytes(c_data)

        callbacks = callbacks_p.contents
        callbacks.private_data.output_reports.append(OutputReport(callbacks.time, data))
        return size

    def _hid_bpf_hw_request(callbacks_p, ctx_p, data_p, size, _type, reqtype):
        if not Callbacks.validate_ctx(callbacks_p, ctx_p):
            return -errno.EINVAL
        DataArray = ctypes.c_uint8 * size
        c_data = DataArray()
        p2 = ctypes.byref(c_data)
        ctypes.memmove(p2, data_p, size)
        req_data = bytes(c_data)
        out_data = bytes(c_data)

        callbacks = callbacks_p.contents
        callbacks.private_data.hw_requests.append(
            HidRawRequest(
                callbacks.time,
                req_data,
                out_data,
                ReportType(_type),
                RequestType(reqtype),
            )
        )
        return size

    def _bpf_map_lookup_elem(callbacks_p, map_p, key_p):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]

        # ensure our type is known
        pdata.bpf.btf.build_struct(map.ctype)

        # allocate / get memory
        map_data_dict = pdata.maps_data.setdefault(map_p, {})
        key = ctypes.c_int.from_address(key_p).value
        data = map_data_dict.setdefault(key, map.ctype())

        # store the returned value in callbacks
        callbacks.helpers_retval = ctypes.cast(ctypes.byref(data), ctypes.c_void_p)

        return 0

    @classmethod
    def _async_init(cls, callbacks_p, async_p, map_p, clock, async_type):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if map_p not in pdata.bpf.maps:
            return -errno.ENOENT

        map = pdata.bpf.maps[map_p]
        map_data_dict = pdata.maps_data[map_p]

        async_field = [
            name for name, value in map.ctype._fields_ if value == async_type
        ]

        # this ensures the provided map_p is correct
        assert len(async_field) == 1

        async_field = async_field.pop()

        for key, value in map_data_dict.items():
            t = ctypes.byref(getattr(value, async_field))
            t_p = ctypes.cast(t, ctypes.c_void_p).value
            if async_p == t_p:
                pdata.asyncs[async_p] = TestAsyncCb(
                    map_p,
                    key,
                    ctypes.cast(ctypes.byref(value), ctypes.c_void_p),
                    None,
                )
                return 0

        return -errno.EINVAL

    def _async_set_callback(callbacks_p, async_p, cb):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if async_p not in pdata.asyncs:
            return -errno.EINVAL

        _async = pdata.asyncs[async_p]
        _async.cb = ctypes.c_void_p(cb)

        return 0

    def _async_start(callbacks_p, async_p, delay, flags):
        callbacks = callbacks_p.contents

        pdata = callbacks.private_data

        if async_p not in pdata.asyncs:
            return -errno.EINVAL

        _async = pdata.asyncs[async_p]

        assert _async.cb is not None

        # store the returned value in callbacks
        callbacks.helpers_retval = ctypes.cast(ctypes.byref(_async), ctypes.c_void_p)

        return 0

    def _bpf_timer_init(callbacks_p, timer_p, map_p, clock):
        return Callbacks._async_init(callbacks_p, timer_p, map_p, clock, BpfTimer)

    def _bpf_wq_init(callbacks_p, wq_p, map_p, clock):
        return Callbacks._async_init(callbacks_p, wq_p, map_p, clock, BpfWq)


@dataclass
class Api:
    """
    Wrapper to make automatically loading functions from a .so file simpler.
    """

    name: str
    args: Tuple[Type[ctypes._SimpleCData | ctypes._Pointer], ...]
    return_type: Optional[Type[ctypes._SimpleCData | ctypes._Pointer]]
    optional: bool

    @property
    def basename(self) -> str:
        return f"_{self.name}"


class Bpf:
    # Cached .so files
    _libs: dict[str, "Bpf"] = {}

    _api_prototypes: list[Api] = [
        Api(
            name="probe",
            args=(ctypes.POINTER(HidProbeArgs),),
            return_type=c_int,
            optional=True,
        ),
        Api(
            name="set_callbacks",
            args=(ctypes.POINTER(Callbacks),),
            return_type=None,
            optional=False,
        ),
    ]

    def __init__(self, lib, btf: Btf, maps: dict[int, Map]):
        self.lib = lib
        self._callbacks = None
        self.maps = maps
        self.btf = btf

    @classmethod
    def _load(cls, name: str) -> Self:
        # Our test setup guarantees this works, running things manually is
        # a bit more complicated.
        ld_path = os.environ.get("LD_LIBRARY_PATH")
        assert ld_path is not None

        sofile = Path(ld_path) / f"{name}.so"
        if not sofile.exists():
            pytest.skip(f"Unable to locate {sofile}, assuming this BPF wasn't built")

        sofile_dir = sofile.with_suffix(".so.p")
        if not sofile_dir.exists():
            pytest.skip(
                f"Unable to locate {sofile_dir}, assuming this BPF wasn't built"
            )

        # We recreate the BTF information for every .so so the Btf class knows
        # about our types
        btf = Btf.load(list(sofile_dir.iterdir()))
        for c in [
            HidProbeArgs,
            HidDevice,
            HidBpfCtx,
            BpfTimer,
            BpfWq,
            TestAsyncCb,
            Callbacks,
        ]:
            btf.build_struct(c)
            assert hasattr(c, "_fields_")

        jsonfile = Path(ld_path) / f"{name}.json"
        if not jsonfile.exists():
            pytest.skip(f"Unable to locate {jsonfile}, assuming this BPF wasn't built")

        # Load the libtest-$BPF.so file first.o, map probe and set_callbacks which
        # have a fixed name.
        #
        # Then try to find the corresponding libtest-$BPF.json file that meson
        # should have generated.
        # Because our actual entry points have custom names we check the json for the
        # right section and then map those we want into fixed-name wrappers, i.e.
        # SEC(HID_BPF_RDESC_FIXUP) becomes self._hid_bpf_rdesc_fixup() which points
        # to the right ctypes function.
        try:
            lib = ctypes.CDLL(sofile.name, use_errno=True)
            assert lib is not None
        except OSError as e:
            pytest.exit(
                f"Error loading the library: {e}. Maybe export LD_LIBRARY_PATH=builddir/test"
            )
        for api in cls._api_prototypes:
            if api.optional and not hasattr(lib, api.name):
                continue
            func = getattr(lib, api.name)
            func.argtypes = api.args
            func.restype = api.return_type
            setattr(lib, api.basename, func)

        maps = {
            ctypes.cast(getattr(lib, m.name), ctypes.c_void_p).value: m
            for m in btf.maps
        }

        try:
            # Only one entry per json file so we're good
            js = json.load(open(jsonfile))[0]
            for program in js["programs"]:

                def register_fun(generic_name):
                    func = getattr(lib, program["name"])
                    func.argtypes = (ctypes.POINTER(HidBpfCtx),)
                    func.restype = c_int
                    setattr(lib, generic_name, func)

                if program["section"].endswith("/hid_bpf_rdesc_fixup") or program[
                    "section"
                ].endswith("/hid_rdesc_fixup"):
                    register_fun("_hid_bpf_rdesc_fixup")
                elif program["section"].endswith("/hid_bpf_device_event") or program[
                    "section"
                ].endswith("/hid_device_event"):
                    register_fun("_hid_bpf_device_event")
        except OSError as e:
            pytest.exit(
                f"Error loading the JSON file: {e}. Unexpected LD_LIBRARY_PATH?"
            )

        return cls(lib, btf, maps)

    @classmethod
    def load(cls, name: str) -> Self:
        """
        Load the given bpf.o file from our tree
        """
        name = f"libtest-{name}"
        if name not in cls._libs:
            cls._libs[name] = cls._load(name)
        instance = cls._libs[name]
        assert instance is not None
        return instance

    def set_callbacks(self, callbacks: Callbacks):
        """
        Set the callbacks to use for the various hid_bpf_* functions that may be
        used by a BPF program. These need to have a matching implementation in
        test-wrapper.c

        For most tests this isn't needed and you can pass the rdesc/report bytes
        directly to hid_bpf_rdesc_fixup() or hid_bpf_device_event().
        """
        self.lib._set_callbacks(callbacks)

    def probe(
        self,
        probe_args: HidProbeArgs,
        private_data: PrivateTestData | None = None,
    ) -> HidProbeArgs:
        """Call the BPF program's probe() function"""
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        callbacks = Callbacks(private_data)
        self.set_callbacks(callbacks)
        # We copy so our caller's probe args are separate from
        # the ones we return after the BPF program modifies them.
        pa = HidProbeArgs()
        p1 = ctypes.byref(probe_args)
        p2 = ctypes.byref(pa)
        ctypes.memmove(p2, p1, ctypes.sizeof(HidProbeArgs))
        pa.hid = callbacks.private_data.id
        rc = self.lib._probe(ctypes.byref(pa))
        if rc != 0:
            raise OSError(rc)
        return pa

    def hid_bpf_device_event(
        self,
        report: bytes | str | None = None,
        private_data: PrivateTestData | None = None,
    ) -> None | bytes:
        """
        Call the BPF program's hid_bpf_device_event function.

        If a report is given, it returns the (possibly modified) report.
        Otherwise it returns None.

        If the report is given with a string representation, it returns
        a string representation as well instead of a byte array.
        """
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        ctx = private_data.current_ctx

        input_is_binary = isinstance(report, bytes) or isinstance(report, bytearray)

        if report is not None:
            binary_report = (
                report
                if input_is_binary
                else binascii.unhexlify(report.replace(" ", ""))
            )
            allocated_size = int(len(binary_report) / 64 + 1) * 64
            data = (ctypes.c_uint8 * allocated_size)(*binary_report)
            callbacks = Callbacks(private_data)
            callbacks.hid_bpf_data = data
            callbacks.hid_bpf_data_sz = allocated_size
            ctx.allocated_size = allocated_size
            ctx.size = len(binary_report)
            self.set_callbacks(callbacks)
        else:
            data = None

        rc = self.lib._hid_bpf_device_event(ctypes.byref(ctx))
        if rc < 0:
            raise OSError(-rc)

        if rc > 0:
            ctx.retval = rc

        if report is None:
            return None
        assert data is not None

        if input_is_binary:
            return bytes(data[: ctx.retval])

        return binascii.hexlify(bytes(data[: ctx.retval]), " ").decode()

    def hid_bpf_rdesc_fixup(
        self,
        rdesc: bytes | str | None = None,
        private_data: PrivateTestData | None = None,
    ) -> None | bytes:
        """
        Call the BPF program's hid_bpf_rdesc_fixup function.

        If an rdesc is given, it returns the (possibly modified) rdesc.
        Otherwise it returns None.

        If the rdesc is given with a string representation, it returns
        a string representation as well instead of a byte array.
        """
        if private_data is None:
            private_data = PrivateTestData(bpf=self)

        ctx = private_data.current_ctx

        input_is_binary = isinstance(rdesc, bytes) or isinstance(rdesc, bytearray)

        if rdesc is not None:
            binary_rdesc = (
                rdesc if input_is_binary else binascii.unhexlify(rdesc.replace(" ", ""))
            )
            allocated_size = 4096
            data = (ctypes.c_uint8 * allocated_size)(*binary_rdesc)
            callbacks = Callbacks(private_data)
            callbacks.hid_bpf_data = data
            callbacks.hid_bpf_data_sz = allocated_size
            ctx.allocated_size = allocated_size
            ctx.size = len(binary_rdesc)
            self.set_callbacks(callbacks)
        else:
            data = None

        rc = self.lib._hid_bpf_rdesc_fixup(ctypes.byref(ctx))
        if rc < 0:
            raise OSError(rc)

        if rc > 0:
            ctx.retval = rc

        if rdesc is None:
            return None
        assert data is not None

        if input_is_binary:
            return bytes(data[: ctx.retval])

        return binascii.hexlify(bytes(data[: ctx.retval]), " ").decode()



#!/usr/bin/env python3
#
# Given the output of udev-hid-bpf inspect, generate a corresponding hwdb file

from dataclasses import dataclass
from itertools import count
from typing import Iterable

import argparse
import json
import sys


@dataclass(order=True)
class Device:
    vid: int
    pid: int
    bustype: int
    group: int
    filename: str


def extract(js) -> Iterable:
    for objfile in js:
        for device in objfile["devices"]:
            yield Device(
                filename=objfile["filename"],
                bustype=int(device["bus"], 16),
                group=int(device["group"], 16),
                vid=int(device["vid"], 16),
                pid=int(device["pid"], 16),
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--prefix", default="HID_BPF_")
    ns = parser.parse_args()
    js = json.load(sys.stdin)
    counter = count()
    devices = sorted(extract(js))
    print("# This file is generated and will be overwritten on updates. Do not edit")
    print("")
    for d in devices:

        def maybe_glob(v):
            return f"{v:04X}" if v != 0 else "*"

        bustype = maybe_glob(d.bustype)
        group = maybe_glob(d.group)
        vid = maybe_glob(d.vid)
        pid = maybe_glob(d.pid)
        print(f"hid-bpf:hid:b{bustype}g{group}v0000{vid}p0000{pid}")
        print(f"  {ns.prefix}{next(counter):03}={d.filename}")
        print(f"  .HID_BPF=1")  # noqa: F541
        print("")
